package de.hft.stuttgart.citydoctor2.datastructure;

import de.hft.stuttgart.citydoctor2.parser.CityGmlParseException;
import de.hft.stuttgart.citydoctor2.parser.CityGmlParser;
import de.hft.stuttgart.citydoctor2.parser.InvalidGmlFileException;
import de.hft.stuttgart.citydoctor2.parser.ParserConfiguration;
import de.hft.stuttgart.citydoctor2.zip.CityGmlZipEntry;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import java.io.Serial;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Reference object for handling of implicit geometries with a library object
 * contained in an external file
 *
 * @author Riegel
 */
public class LibraryObject extends Geometry {

	private static final Logger logger = LogManager.getLogger(LibraryObject.class);
	@Serial
	private static final long serialVersionUID = -50293435187454911L;

	private static Map<String, LibraryObject> libraryObjects = new ConcurrentHashMap<>();

    public static LibraryObject of(Path path, ParserConfiguration config) {
        if (libraryObjects.containsKey(path.toString())) {
            return libraryObjects.get(path.toString());
        }
        Geometry protoGeom = parseFile(path, config);
        if (protoGeom == null) {
            return null;
        }
        LibraryObject libOb = new LibraryObject(protoGeom.getType(), protoGeom.getLod());
        protoGeom.getPolygons().forEach(libOb::addPolygon);
        libOb.updateEdgesAndVertices();
        libraryObjects.put(path.toString(), libOb);
        return libOb;
    }

    public static LibraryObject of(CityGmlZipEntry entry, ParserConfiguration config) {
        String fileName = entry.getEntrySubPath();
        if (libraryObjects.containsKey(fileName)){
            return libraryObjects.get(fileName);
        }
        Geometry protoGeom = parseZipEntry(entry, config);
        if (protoGeom == null) {
            return null;
        }
        LibraryObject libOb = new LibraryObject(protoGeom.getType(), protoGeom.getLod());
        protoGeom.getPolygons().forEach(libOb::addPolygon);
        libOb.updateEdgesAndVertices();
        libraryObjects.put(fileName, libOb);
        return libOb;
    }

    private LibraryObject(GeometryType type, Lod lod) {
        super(type, lod);
    }


    private static Geometry parseFile(Path path, ParserConfiguration config) {
        Geometry geo = null;
        if (Files.exists(path)) {
            try {
                geo = getProtoGeometry(CityGmlParser.parseCityGmlFileSilently(path.toString(), config));
            } catch (CityGmlParseException e) {
                logger.error(String.format(
                        "Encountered an error while parsing library object %s", path));
                logger.error(e.getStackTrace());
            } catch (InvalidGmlFileException e) {
                logger.error(e.getStackTrace());
            }
        } else {
            logger.error(String.format("Implicit geometry references non-existing library object file %s.", path));
        }
        return geo;
    }

    private static Geometry parseZipEntry(CityGmlZipEntry entry, ParserConfiguration config) {
        Geometry geo = null;
        try {
            entry.loadEntry(config);
            geo = getProtoGeometry(entry.getModel());
        } catch (InvalidGmlFileException e) {
            logger.error(e.getStackTrace());
        } catch (CityGmlParseException e) {
            logger.error(String.format(
                    "Encountered an error while parsing library object %s", entry.getEntrySubPath()));
            logger.error(e.getStackTrace());
        }
        return geo;
    }

    private static Geometry getProtoGeometry(CityDoctorModel model) throws InvalidGmlFileException, CityGmlParseException {
        if (model == null) {
            throw new CityGmlParseException("CityDoctorModel of referenced LibraryObject is null");
        }
        List<CityObject> objects = model.createFeatureStream().toList();
        if (objects.isEmpty()) {
            throw new InvalidGmlFileException("Referenced library-object gml file does not contain any CityGML objects!");
        } else if (objects.size() > 1) {
            throw new InvalidGmlFileException("Referenced library-object gml file contains more than one CityGML object!");
        }
        return objects.get(0).getHighestLodGeometry();
    }

}
